/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * JBIListener.java - Initial Developper : ofabre
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.validation.listener;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.Hashtable;
import java.util.Map;
import java.util.logging.Level;

import javax.xml.XMLConstants;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.performance.PerformanceNotification;
import org.ow2.petals.se.validation.Constants;
import org.ow2.petals.se.validation.service.ValidationService;
import org.ow2.petals.se.validation.util.XsdResourceResolver;
import org.xml.sax.SAXException;

/**
 * The VALIDATION service engine JBIListener
 * 
 * @author Adrien LOUIS - EBMWebSourcing
 * @author mbeldame
 * 
 */
public class JBIListener extends AbstractJBIListener {

    private static Map<String, Schema> SCHEMA_CACHE = new Hashtable<String, Schema>();

    private ValidationService validationService;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.component.framework.listener.AbstractListener#init()
     */
    @Override
    protected void init() {
        this.validationService = new ValidationService(getLogger());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.component.framework.listener.AbstractJBIListener
     * #onJBIMessage(org.objectweb.petals.component.framework.util.Exchange)
     */
    @Override
    protected void onJBIMessage(final Exchange exchange) {
        if (exchange.isActiveStatus() && exchange.getFault() == null) {
            if (exchange.isInOutPattern()) {
                this.processExchange(exchange);
            } else {
                // refuse the others patterns than InOut
                this.getLogger().warning(
                        "Exchange pattern (" + exchange.getPattern() + ") not supported");
                exchange.setError(new Exception("Unsupported exchange pattern"));
            }
        }
    }

    /**
     * Retrieve the Schema associated to this XSD, and update the corresponding
     * entry in the cache. Identifier is composed of a String which value is the
     * location of the XSD file.
     * 
     * @param xsdResourceInputStream
     * @param identifier
     * @return
     * @throws SAXException
     *             Failed to create the XSD Schema
     */
    private Schema getSchemasAndUpdateCache(final InputStream xsdResourceInputStream,
            final String identifier) throws SAXException {
        Schema result = null;

        if (xsdResourceInputStream == null) {
            return null;
        }

        final Source source = new StreamSource(xsdResourceInputStream);

        final SchemaFactory schemaFactory = SchemaFactory
                .newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
        final XsdResourceResolver res = new XsdResourceResolver(this.getInstallRoot(),getLogger());
        schemaFactory.setResourceResolver(res);

        result = schemaFactory.newSchema(source);
        SCHEMA_CACHE.put(identifier, result);

        this.getLogger().log(Level.INFO, "XSD stored in cache. Identifier:" + identifier);

        return result;
    }

    /**
     * Retrieve the InputStream associated to the desired xsd from the Consumes.
     * 
     * @return
     */
    private Schema getXsdSchema() {
        this.getLogger().log(Level.INFO, "Retrieve the XSD schema to use :");
        final String xsdResourcePath = this.getExtensions().get(Constants.XSD_FILE);
        this.getLogger().log(Level.INFO, "Retrieve the XSD schema to use :" +xsdResourcePath);
        
        Schema xsdSchema = JBIListener.SCHEMA_CACHE.get(this.getInstallRoot() + xsdResourcePath);

        if (xsdSchema == null) {

            InputStream xsdStream;
            // try to get the XSD from the filesystem
            try {
                xsdStream = new FileInputStream(new File(this.getInstallRoot(), xsdResourcePath));
            } catch (final FileNotFoundException e) {
                throw new RuntimeException("Failed to retrieve XSD resource - "
                        + e.getClass().getSimpleName() + " - " + e.getMessage());
            }

            try {
                xsdSchema = this.getSchemasAndUpdateCache(xsdStream, this.getInstallRoot()
                        + xsdResourcePath);
            } catch (final SAXException e) {
                throw new RuntimeException("Failed to build XSD schemas cache - "
                        + e.getClass().getSimpleName() + " - " + e.getMessage());
            }

        }
        return xsdSchema;
    }

    /**
     * Process the transformation and set the response depending on the
     * operation called: validate or other.
     * 
     * @param exchange
     */
    private void processExchange(final Exchange exchange) {
        try {
            final PerformanceNotification performanceNotification = this.component
                    .createPerformanceNotificationSendingToOutside("Time for XML Validation",
                            exchange);

            // get the XSD document for current Service Unit
            final Schema xsdSchema = this.getXsdSchema();
            if (xsdSchema == null) {
                exchange.setError(new RuntimeException(
                        "No XSD Schema is found to transform the document."));
            } else {

                // Validate the input depending on the operation
                // general behavior: (operation == validate) i.e. validate mode
                if (exchange.getOperation() != null
                        && Constants.VALIDATION_OPERATION_VALIDATE.equals(exchange.getOperation()
                                .getLocalPart())) {
                    validationService.validate(exchange, xsdSchema);
                }
                // operation != validate i.e. filter mode
                else {
                    validationService.filter(exchange,xsdSchema);
                }
            }
            this.component.sendPerformanceNotification(performanceNotification);

        } catch (final Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException) e;
            } else {
                throw new RuntimeException(e.getClass().getSimpleName() + " - " + e.getMessage());
            }
        }
    }
}
