/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.validation.service;

import java.io.IOException;
import java.util.Set;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.security.auth.Subject;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.Validator;

import org.apache.commons.lang.CharEncoding;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.util.InputStreamForker;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.se.validation.Constants;
import org.ow2.petals.se.validation.ValidationException;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The VALIDATION service
 * 
 * @author Adrien LOUIS - EBMWebSourcing
 * @author mbeldame
 * 
 */
public class ValidationService {

    private Logger logger;

    public ValidationService(Logger logger) {
        super();
        this.logger = logger;
    }

    /**
     * Validate operation : if message is valid, TRUE is returned, otherwise
     * FALSE and a Comment are returned
     * 
     * @param exchange
     */
    public void validate(Exchange exchange, Schema xsdSchema) {
        try {

            // get the XSD document for current Service Unit
            if (xsdSchema == null) {
                exchange
                        .setError(new Exception("No XSD Schema is found to transform the document."));
            } else {

                // Get in Message content
                final Source inContent = exchange.getInMessageContentAsSource();

                // Validation
                xmlValidation(inContent, xsdSchema.newValidator());
                logger.info("XML message is valid.");
                final Subject subject = exchange.getInMessage().getSecuritySubject();
                try {
                    exchange.setOutMessageContent(SourceHelper.createSource(
                            Constants.VALID_XML_RESPONSE, CharEncoding.UTF_8));
                    exchange.setOutSubject(subject);
                } catch (PEtALSCDKException e) {
                    exchange.setError(e);
                }

            }
        } catch (final ValidationException xmle) {

            logger.info("XML message is invalid." + xmle.getMessage());
            try {
                final Subject subject = exchange.getInMessage().getSecuritySubject();
                exchange.setOutMessageContent(SourceHelper.createSource(
                        Constants.INVALID_XML_RESPONSE.replace("$reason", xmle.getMessage()),
                        CharEncoding.UTF_8));
                exchange.setOutSubject(subject);
            } catch (MessagingException e) {
                exchange.setError(e);
            } catch (PEtALSCDKException e) {
                exchange.setError(e);
            }
        } catch (MessagingException e) {
            exchange.setError(e);
        }

    }

    /**
     * Filter operation : if message is valid, it is forwarded on the OUT
     * Otherwise, a Fault is returned
     * 
     * @param exchange
     * @param xsdSchema
     */
    public void filter(Exchange exchange, Schema xsdSchema) {
        InputStreamForker messageAsStream = null;
        try {
            if (exchange.getInMessageContentAsSource() instanceof StreamSource) {
                messageAsStream = new InputStreamForker(((StreamSource) exchange
                        .getInMessageContentAsSource()).getInputStream());

                final Source source = new StreamSource(messageAsStream.getInputStreamOne());

                try {
                    xmlValidation(source, xsdSchema.newValidator());
                    final Source source2 = new StreamSource(messageAsStream.getInputStreamTwo());
                    logger.info("XML message is valid.");
                    final Subject subject = exchange.getInMessage().getSecuritySubject();
                    exchange.setOutMessageContent(source2);
                    exchange.setOutSubject(subject);
                    forwardAttachments(exchange);
                    forwardProperties(exchange);
                } catch (final ValidationException xmle) {
                    logger.info("XML message is invalid." + xmle.getMessage());
                    try {
                        setValidationFaultOnExchange(exchange, xmle.getMessage());
                    } catch (JBIException e) {
                        exchange.setError(e);
                    }
                }
            }
        } catch (MessagingException e) {
            exchange.setError(e);
        }
    }

    public void xmlValidation(Source xml, Validator validator) throws ValidationException {
        try {
            validator.validate(xml);
        } catch (final SAXParseException e) {
            throw new ValidationException(e.getMessage());
        } catch (final SAXException e) {
            throw new ValidationException("Failed to validate message - "
                    + e.getClass().getSimpleName() + " - " + e.getMessage());
        } catch (final IOException e) {
            throw new ValidationException("Failed to validate message - "
                    + e.getClass().getSimpleName() + " - " + e.getMessage());
        }
    }

    /**
     * move IN attachments to OUT attachments
     * 
     * @param exchange
     * @throws MessagingException
     */
    @SuppressWarnings("unchecked")
    private void forwardAttachments(final Exchange exchange) throws MessagingException {
        Set<String> names = exchange.getInMessage().getAttachmentNames();
        for (String name : names) {
            logger.finest("Forward " + name + " attachment on the response.");
            DataHandler handler = exchange.getInMessage().getAttachment(name);
            exchange.getOutMessage().addAttachment(name, handler);
        }
    }

    /**
     * move IN properties to OUT properties
     * 
     * @param exchange
     * @throws MessagingException
     */
    @SuppressWarnings("unchecked")
    private void forwardProperties(final Exchange exchange) throws MessagingException {
        Set<String> names = exchange.getInMessage().getPropertyNames();
        for (String name : names) {
            logger.finest("Forward " + name + " property on the response.");
            Object prop = exchange.getInMessage().getProperty(name);
            exchange.getOutMessage().setProperty(name, prop);
        }
    }

    private void setValidationFaultOnExchange(Exchange exchange, String message)
            throws JBIException {
        StringBuffer faultMessage = new StringBuffer();
        faultMessage.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        faultMessage.append("<tns:validationFault xmlns:tns='");
        faultMessage.append(Constants.VALIDATION_SERVICE_NS);
        faultMessage.append("'><tns:message>");
        faultMessage.append(message);
        faultMessage.append("</tns:message></tns:validationFault>");
        Fault fault = exchange.createFault();
        try {
            fault
                    .setContent(SourceHelper.createSource(faultMessage.toString(),
                            CharEncoding.UTF_8));
            exchange.setFault(fault);
        } catch (PEtALSCDKException e) {
            throw new JBIException(e.getMessage());
        }
    }
}
